//! Marathon Wad format handling.

use std::collections::BTreeMap;
use std::fmt;

use durandal::bin::*;
use durandal::machead::try_mac_header;
use durandal::text::mac_roman_conv;

type Chunk   <'a> = &'a[u8];
type ChunkMap<'a> = BTreeMap<Ident, Chunk<'a>>;
type EntryMap<'a> = BTreeMap<u16  , Entry<'a>>;

pub struct Entry<'a>
{
   pub map: ChunkMap<'a>,
   pub ext: &'a[u8],
}

#[derive(Debug)]
pub struct Wad<'a>
{
       ver: Ver,
       dvr: u16,
       ext: usize,
       nam: String,
   pub ent: EntryMap<'a>,
}

#[derive(Debug)]
pub enum Ver
{
   MI,
   M2HasOvr,
   M2HasDir,
   M1
}

impl<'a> fmt::Debug for Entry<'a>
{
   fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result
   {
      write!(f, "Entry {{ ")?;
      for (k, _) in &self.map {write!(f, "{} ", mac_roman_conv(&k[..]))?}
      write!(f, "}}")
   }
}

impl<'a> Wad<'a>
{
   pub fn new(b: &[u8]) -> Wad
   {
      let   b = &b[try_mac_header(b)..];
      let ver = b_u16b(&b[  .. 2]);
      let dvr = b_u16b(&b[ 2.. 4]);
      let nam =        &b[ 4..68];
      //  crc = b_u32b(&b[68..72]);
      let dir = b_u32b(&b[72..76]) as usize;
      let num = b_u16b(&b[76..78]) as usize;
      let ext = b_u16b(&b[78..80]) as usize;
      //  hdr = b_u16b(&b[80..82]);
      //  bsz = b_u16b(&b[82..84]);
      //  pck = b_u32b(&b[84..88]);

      let ver = match ver {
      4 => Ver::MI,
      2 => Ver::M2HasOvr,
      1 => Ver::M2HasDir,
      0 => panic!("old wad type unsupported"), //Ver::M1,
      _ => panic!("invalid wad version"),
      };

      let mut map = EntryMap::new();
      let mut   p = dir;
      let       n = match ver {Ver::M1 | Ver::M2HasDir => true, _ => false};
      let       h = if n {8} else {10};

      for i in 0..num
      {
         let ofs =       b_u32b(&b[p  ..p+ 4]) as usize;
         let len =       b_u32b(&b[p+4..p+ 8]) as usize;
         let ind = if n {i as u16}
                   else {b_u16b(&b[p+8..p+10])};
         let ent = Entry{map: get_chunks(&b[ofs..ofs+len]),
                         ext:            &b[p+h..p+h+ext]};

         map.insert(ind, ent);

         p += h + ext;
      }

      Wad{ver, dvr, ext,
          nam: mac_roman_conv(nam),
          ent: map}
   }
}

fn get_chunks(b: &[u8]) -> ChunkMap
{
   let mut   p = 0;
   let mut map = ChunkMap::new();

   while p < b.len()
   {
      let k = b_iden(&b[p   ..p+ 4]);
      // nx = b_u32b(&b[p+ 4..p+ 8]);
      let l = b_u32b(&b[p+ 8..p+12]) as usize;
      //  o = b_u32b(&b[p+12..p+16]);
      map.insert(k,  &b[p+16..p+16+l]);
      p += l + 16;
   }

   map
}

// EOF
